package mic1.gui;
/*
 * @(#)MemoryTable.java	1.00 18/Jan/2001
 * 
 *  This program is free software; you can redistribute it and/or modify 
 *  it under the terms of the GNU General Public License as published by 
 *  the Free Software Foundation; either version 2 of the License, or 
 *  (at your option) any later version. 
 *
 * @Version : 1.00 18/Jan/2001
 * @Author  : Richard Klos
 *
 * Description : A set of tables which show the memory adresses and the
 *               memory contents of these addresses.
 * 
***  

To Do:


Modification History:

Name             Date       Comment
---------------- ---------- ----------------------------------------
Richard Klos     18.01.2001 Created.
Pierre Bakker	 22.01.2001 Ajusted the minimal width                      
Pierre Bakker	 22.01.2001 Cleaned up
Richard Klos     03.10.2001 Added support for 32 bits view (WORD addresses).
Tim Helsloot     12.03.2003 32 bits support wasn`t OK, corrected
*/

import mic1.ijvm.Mic1Interface;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import javax.swing.table.DefaultTableModel;
import javax.swing.JPanel;
import javax.swing.JScrollBar;


class MemoryTable extends JPanel implements AdjustmentListener, ComponentListener
{
	// The two columns.
	private ResizableColumn m_addresses,
	                        m_values;
	// The memory range.
	private int m_rangeStart, m_rangeEnd;

	private Mic1Interface m_mic1;

	// The adjustable.
	private Adjustable m_adjustable;

	// Flag indicating a reqeust for a update
	// caused by an adjustment.
	private boolean m_updateByAdjustment;

	// Previous offset.
	private int m_prevOffset;
	
	// The size of the element (in this class either 1 or 4).
	private int m_elementSize;
	
	
	public MemoryTable(Mic1Interface mi, int rangeStart, int rangeEnd, boolean intView)
	{
		m_mic1       = mi;
		if(intView)
			m_elementSize = 4;
		else
			m_elementSize = 1;

		m_rangeStart = rangeStart / m_elementSize;
		m_rangeEnd   = rangeEnd / m_elementSize;
		
		m_addresses = new ResizableColumn();
		m_values    = new ResizableColumn();

		m_updateByAdjustment = true;
		m_prevOffset = 0;

		(m_addresses.getColumnModel().getColumn(0)).setMinWidth(75);
		m_addresses.setShowGrid(false);
		m_addresses.setBackground(Color.lightGray);
		m_addresses.setForeground(Color.blue);
		m_addresses.setEnabled(false);				

		(m_values.getColumnModel().getColumn(0)).setMinWidth(75);
		(m_values.getColumnModel().getColumn(0)).setMaxWidth(75);
		m_values.setEnabled(false);

		JScrollBar scrollbar = new JScrollBar(JScrollBar.VERTICAL);
		m_adjustable = scrollbar;
		m_adjustable.addAdjustmentListener(this);
		m_adjustable.setUnitIncrement(1);
		m_adjustable.setMinimum(m_rangeStart);
		m_adjustable.setMaximum(m_rangeEnd);
			
		setLayout(new BorderLayout(0,0));
		add(m_addresses, BorderLayout.WEST);
		add(m_values, BorderLayout.CENTER);
		add(scrollbar, BorderLayout.EAST);

		addComponentListener(this);
	}
	
	
	/**
	 * Adds a AdjustmentListener to the scrollbar
	 * of this MemoryTable.
	 */
	public void addAdjustmentListener(AdjustmentListener al)
	{
		m_adjustable.addAdjustmentListener(al);
	}
	
	
	/**
	 * Returns the address at the top of the
	 * visible part of this table.
	 */
	public int getTopAddress()
	{
		return m_adjustable.getValue();
	}
	

	/**
	 * Returns the address at the bottom of the
	 * visible part of this table.
	 */
	public int getBottomAddress()
	{
		return m_adjustable.getValue() + m_addresses.getRowCount();
	}
	


	/**
	 * Centers the view on the specified address.
	 *
	 * @param: address
	 */	
	public void centerOn(int address)
	{
		int newOffset = validateOffset(address - getRowCount() / 2);
		m_adjustable.setValue(newOffset);
	}
	
	
	/**
	 * Returns the pixel height of a row.
	 */
	public int getRowHeight()
	{
		return m_addresses.getRowHeight();
	}
	
	
	/**
	 * Returns the number of rows.
	 */
	public int getRowCount()
	{
		return m_addresses.getRowCount();
	}
	
	
	/**
	 * Writes the table with memory addresses and their contents.
	 */
	public void updateView()
	{
		int offset = validateOffset(m_adjustable.getValue());

		// If update called by scrollbar.
		if(m_updateByAdjustment)
		{
			m_updateByAdjustment = false;
			// If offset has not changed.
			if(offset == m_prevOffset)
				return;
			else
				m_prevOffset = offset;
		}
		
		for(int i = getRowCount() - 1; i >= 0; i--)
		{
			int address = (offset + i);
			if(address <= m_rangeEnd  &&  address >= m_rangeStart)
			{
				address *= m_elementSize;
				m_addresses.setValueAt(mic1.util.Util.toHex(address, 8), i, 0);
				int integer = 0;
				for(int jj = m_elementSize - 1; jj >= 0 ; jj--)
				{
					integer |= m_mic1.getMemoryContent(address + jj) << ((m_elementSize - (jj + 1)) * 8);
				}
				m_values.setValueAt(mic1.util.Util.toHex(integer, 2 * m_elementSize), i, 0);
			}		
			else
			{
				m_addresses.setValueAt(".", i, 0);
				m_values.setValueAt(".", i, 0);
			}
		}
	}
	
	
	/**
	 * Determans if the offset lies between or 
	 * outside the visible rows.
	 *
	 * @param: offset
	 */
	private int validateOffset(int offset)
	{
		if(offset < m_rangeStart)
			return m_rangeStart;
		else if(offset > m_rangeEnd - getRowCount() + 1)
			return offset = m_rangeEnd - getRowCount() + 1;
		else
			return offset;
	}


	// AdjustmentListener
	public void adjustmentValueChanged(AdjustmentEvent e)
	{
		m_updateByAdjustment = true;
		updateView();
	}
	
	
	// ComponentListener
	public void componentHidden(ComponentEvent e) {}
	public void componentMoved(ComponentEvent e) {}
	public void componentShown(ComponentEvent e) {}
	public void componentResized(ComponentEvent e)
	{
		// Resize columns.
		m_addresses.resize(getHeight());
		m_values.resize(getHeight());
		
		// Change block increment to 90% of the visible range.
		m_adjustable.setBlockIncrement((int)(0.5 + getRowCount() * 0.9));
		
		updateView();
	}
}
