/**
 * Gutter.java
 * Copyright (C) 1999, 2000 mike dillon
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/**
	Modification History:
	
	Name			 Date	    Comment
	---------------- ---------- -----------------------------------------------------
	Andy Chu		 22.11.2000 Deleted code which is not needed for the MIC1 project
	Andy Chu		 25.12.2000 Linenumbers activated
	Andy Chu		 04.01.2001 Breakpoints added
 */


package mic1.gui.editor.textarea;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.Border;
import mic1.ijvm.*;
import mic1.gui.editor.*;
import mic1.gui.*;

public class Gutter extends JComponent implements SwingConstants
{
	// private members
	// the JEditTextArea this gutter is attached to
	private int[] breakpoints;
	private JEditTextArea textArea;

	private int baseline = 0;
	private int ileft = 0;

	private Dimension gutterSize = new Dimension(0,0);
	private Dimension collapsedSize = new Dimension(0,0);

	private FontMetrics fm;

	private boolean collapsed = true;

	private Font font;
	private int extraWidth;
	private int extraSpace;
	private Mic1Interface m_Simulator;
	private MainFrame m_Parent;
	private PropertiesManager m_PropertiesManager;

	public Gutter(JEditTextArea textArea, TextAreaDefaults defaults)
	{
		this(textArea, defaults, null);		
	}
	
	public Gutter(JEditTextArea textArea, TextAreaDefaults defaults, int[] _breakpoints)
	{
		this.textArea = textArea;
		
		setBackground(defaults.lineNrBgColor);
		setForeground(defaults.lineNrColor);

		font = textArea.getPainter().getFont();
		extraWidth = font.getSize()/2;
		
		breakpoints = _breakpoints;
		setBorder(0, defaults.lineNrBgColor);
		setGutterWidth(1);
		setLineNrEnabled(false);
	}

	/*
	 * Set the breakpoints array;
	 */
	protected void setBreakpoints(int[] bp)
	{
		breakpoints = bp;
		repaint();
	}

	public void paintComponent(Graphics gfx)
	{
		if (!collapsed)
		{
			// fill the background
			Rectangle r = gfx.getClipBounds();
			gfx.setColor(getBackground());
			gfx.fillRect(r.x, r.y, r.width, r.height);

			paintLineNumbers(gfx);
		}
	}

	protected void paintLineNumbers(Graphics gfx)
	{	
		FontMetrics pfm = textArea.getPainter().getFontMetrics();
		int lineHeight = pfm.getHeight();
		int baseline = (int) Math.round((this.baseline + lineHeight -
			pfm.getMaxDescent()) / 2.0);

		int firstLine = textArea.getFirstLine() + 1;
		int lastLine = firstLine + (getHeight() / lineHeight);

		int firstValidLine = (int) Math.max(1, firstLine);
		int lastValidLine = (int) Math.min(textArea.getLineCount(),
			lastLine);
			
		gfx.setColor(getForeground());

		String number;

		for (int line = firstLine; line <= lastLine;
			line++, baseline += lineHeight)
		{
			// only print numbers for valid lines
			if (line < firstValidLine || line > lastValidLine)
				continue;

			number = Integer.toString(line);

			int offset = 2;

			if(breakpoints[0] != 0)
			{
				extraSpace = pfm.charWidth('*');
				extraWidth = font.getSize();
			}				
			else
			{
				extraSpace = 0;
				extraWidth = font.getSize()/2;
			}				
			
			offset = gutterSize.width - collapsedSize.width
				- (pfm.stringWidth(number) + 1) - extraSpace;

			if (isHighlighted(line))
			{
				int xOffset = ileft + offset;
				gfx.drawString(number, xOffset, baseline);
				
				Polygon p = new Polygon();
				int baseOffset = xOffset+pfm.stringWidth(number);
				p.addPoint(baseOffset, baseline-pfm.getHeight()/2);
				p.addPoint(baseOffset, baseline);
				p.addPoint(baseOffset+pfm.charWidth('*'), baseline-pfm.getHeight()/4);
				
				gfx.setColor(Color.cyan);
				gfx.fillPolygon(p);
				
				gfx.setColor(Color.red);				
				gfx.drawPolygon(p);
				
				gfx.setColor(getForeground());
			}
			else
			{
				gfx.setColor(getForeground());
				gfx.drawString(number, ileft + offset, baseline);
			}
		}
		setGutterWidth(pfm.stringWidth(Integer.toString(lastValidLine)) + extraWidth);
	}

	/**
	 * Convenience method for setting a default matte border on the right
	 * with the specified border width and color
	 * @param width The border width (in pixels)
	 * @param color The border color
	 */
	public void setBorder(int width, Color color)
	{
		setBorder(BorderFactory.createMatteBorder(0,0,0,width,color));
	}

	/*
	 * JComponent.setBorder(Border) is overridden here to cache the left
	 * inset of the border (if any) to avoid having to fetch it during every
	 * repaint.
	 */
	public void setBorder(Border border)
	{
		super.setBorder(border);

		if (border == null)
		{
			ileft = 0;
			collapsedSize.width = 0;
			collapsedSize.height = 0;
		}
		else
		{
			Insets insets = border.getBorderInsets(this);
			ileft = insets.left;
			collapsedSize.width = insets.left + insets.right;
			collapsedSize.height = insets.top + insets.bottom;
		}
	}

	/*
	 * JComponent.setFont(Font) is overridden here to cache the baseline for
	 * the font. This avoids having to get the font metrics during every
	 * repaint.
	 */
	public void setFont(Font font)
	{		
		super.setFont(font);
		fm = getFontMetrics(font);
		baseline = fm.getHeight() - fm.getMaxDescent();
		repaint();
	}

	/**
	 * Set the width of the expanded gutter
	 * @param width The gutter width
	 */
	public void setGutterWidth(int width)
	{
		if (width < collapsedSize.width) width = collapsedSize.width;

		gutterSize.width = width;		

		// if the gutter is expanded, ask the text area to revalidate
		// the layout to resize the gutter
		if (!collapsed) textArea.revalidate();
	}

	/**
	 * Get the width of the expanded gutter
	 * @return The gutter width
	 */
	public int getGutterWidth()
	{
		return gutterSize.width;
	}

	/*
	 * Component.getPreferredSize() is overridden here to support the
	 * collapsing behavior.
	 */
	public Dimension getPreferredSize()
	{
		if (collapsed)
		{
			return collapsedSize;
		}
		else
		{
			return gutterSize;
		}
	}

	public Dimension getMinimumSize()
	{
		return getPreferredSize();
	}

	/**
	 * Turns the line numbering on or off and causes the gutter to be
	 * repainted.
	 * @param enabled true if line numbers are drawn, false otherwise
	 */
	public void setLineNrEnabled(boolean enabled)
	{
		if (collapsed != enabled) return;

		collapsed = !enabled;

		textArea.revalidate();
	}


	private boolean isHighlighted(int line)
	{
		if(breakpoints == null)return false;
		
		for (int i = 0; i < breakpoints.length; i++)
		{
			if (line == breakpoints[i])
				return true;
		}
		return false;
	}
}