package mic1.ijvm;
/*
 * @(#)Mic1Controller.java	1.0 02/Nov/2000
 * 
 *  This program is free software; you can redistribute it and/or modify 
 *  it under the terms of the GNU General Public License as published by 
 *  the Free Software Foundation; either version 2 of the License, or 
 *  (at your option) any later version. 
 *
 * @Version : 1.00 02/Nov/2000
 * @Author  : WDW
 * @Dependencies : Mic1Output
 *
 * Description : 
 * 
 * Input interface for controlling the MIC1 machine.
 * 

Modification History:

Name             Date       Comment
---------------- ---------- ----------------------------------------
WDW              02.11.2000 Created
Richard Klos     06.12.2000 Added method onConsoleInput(char)
Richard Klos     16.12.2000 Changed method name onConsoleInput
                            to sendChar.
Pierre Bakker	 19.12.2000 added method refreshOutput()
                            

*/
 
 
import java.io.*;

public interface Mic1Interface
{
	/**
	 * Loads a new program in memory.
	 *
	 * @param : filepath of the program
	 * @exception : MicException if there is still an active program.
	 * @exception : IOException for file access;
	 */	
	public void loadProgram (String sFileName) throws Exception;


	/** 
	 * Returns true when there is valid program running/loaded.
	 *
	 * @return : boolean!
	 */	
	public boolean validProgram ();
													 
	/** 
	 * Returns the filepath of the active program file
	 *
	 * @return : Empty String when there is no program active.
	 */	
	public String getProgramFile ();
	
	
	/** 
	 * Starts running the current program. Breakpoints are skipped
	 *
	 * @exception : MicException if there is no program loaded.
	 * @exception : MicException if there is a running program.
	 */	
	public void run () throws Exception;
	
	
	/** 
	 * Start or Resumes the current program until the next breakpoint.
	 *
	 * @exception : MicException if there is no program loaded.
	 * @exception : MicException if there is a running program.
	 */		
	public void runToBreakPoint () throws Exception;
	

	/** 
	 * Executes the current instruction and increments the PC.
	 *
	 * @return : Returns false when end of program is reached.
	 * @exception : MicException if there is no program loaded.
	 * @exception : MicException if there is a running program.
	 */	
	public boolean step () throws Exception;
	
	
	/** 
	 * Resets the machine. The program loaded persists in memory
	 *
	 * @exception : MicException if there is a running program.
	 */		
	public void reset () throws Exception;
	
	/** 
	 * Forces a Kernal Resets of the machine. The program loaded is cleared from memory
	 *
	 * @exception : MicException if there is a running program.
	 */		
	public void unloadProgram () throws Exception;

	/** 
	 * Stops the current program if running.
	 *
	 */			
	public void stop ();

	/** 
	 * Determines if the program is running.
	 *
	 * @return : true when running!
	 */	
	public boolean isRunning ();

	/** 
	 * Adds a new linenumber to list overlapping the old one. 
	 *
	 * @param : the linenumber of the breakpoint.
	 * @exception : MicException if there is a running program.
	 */		
	public void setBreakPoint (int nLineNumber) throws Exception;
	
	/** 
	 * Removes a linenumber from the list if found.
	 *
	 * @param : the linenumber of the breakpoint. 
	 * @exception : MicException if there is a running program.
	 * @return : false if the breakpoint doesn't exit
	 */	
	public boolean removeBreakPoint (int nLineNumber) throws Exception;
	
	/** 
	 * Validates the linenumbers of the breakpoint list. 
	 * Removing and returning the invalid ones.
	 *
	 * @exception : MicException if there is no program loaded.
	 * @exception : MicException if there is a running program.
	 */		
	public int[] validateBreakPoints () throws Exception;
		
	/** 
	 * Removes all breakpoints.
	 *
	 * @exception : MicException if there is a running program.
	 */		
	public void clearBreakPoints () throws Exception;
	
	/** 
	 * Fills and returns an array containing the breakpoint.
	 * It's not a reference to the internal list.
	 * 
	 * @return : int array containing the linenumbers.
	 */			
	public int[] getBreakPoints ();

	/** 
	 * The interface which is used by the machine for output messages
	 *
	 */			
	public void setOutputInterface (Mic1OutputInterface output);
	

	/////////////////////////////////////////////////////////////////////
	// Setting register values
	/////////////////////////////////////////////////////////////////////
	
	/** 
	 * Memory address register.
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */	
	public void setRegisterMAR (int nValue) throws Exception;
	
	/** 
	 * Register that contains word values for writing or reading 
	 * from main memory
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */	
	public void setRegisterMDR (int nValue) throws Exception;
	
	/** 
	 * Program Counter
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */		
	public void setRegisterPC (int nValue) throws Exception;
	
	/** 
	 * Register that stores bytes fetched from memory.
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */		
	public void setRegisterMBR (int nValue) throws Exception;
	
	/** 
	 * Top point of the current stackframe
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */	
	public void setRegisterSP (int nValue) throws Exception;
	
	/** 
	 * Bottom point of the current stackframe
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */		
	public void setRegisterLV (int nValue) throws Exception;
	
	/** 
	 * Points to the const pool
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */		
	public void setRegisterCPP (int nValue) throws Exception;
	
	/** 
	 * ??
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */		
	public void setRegisterTOS (int nValue) throws Exception;
	
	/** 
	 * scratch register
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */		
	public void setRegisterOPC (int nValue) throws Exception;

	/** 
	 * ??
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */		
	public void setRegisterH (int nValue) throws Exception;
	
	/** 
	 * Contains the address of the next microinstruction.
	 *
	 * @param : Hex value.
	 * @exception : MicException if there is a running program.
	 */		
	public void setRegisterMPC (int nValue) throws Exception;
	
	/**
	 * Sends a character to the machines keyboard buffer.
	 *
	 * @param : a character.
	 */
	public void sendChar(char character);
	
	/**
	 * Returns the byte at the specified memory address.
	 * If the address is out of range, this method
	 * returns 255 (0xFF), like in most real computer
	 * systems.
	 *
	 * @param : memory address.
	 */
	public int getMemoryContent(int address);
	
	/**
	 * Send a refresh signal to the machine
	 */
	public void refreshOutput();
	
};

/*EOF*/