package mic1.ijvm;


/*

Modification History:

Name             Date       Comment
---------------- ---------- ----------------------------------------
WDW              02.11.2000 Created
Richard Klos     06.12.2000 Added method onConsoleInput(char)
Richard Klos     16.12.2000 Changed method name onConsoleInput
                            to sendChar.
                            Added method getMemoryContent(int).
Richard Klos     16.01.2001 Removed all hex-strings as parameters to
                            functions to the Mic1OutputInterface.
Richard Klos     03.10.2001 Small modifications to use improved MemoryView.
Richard Klos     04.10.2001 m_Clock is made null when running stops.
Tim Helsloot     12.03.2003 LV, SP and CPP: word addresses, MAR: byte address.
*/

import java.io.*;
import java.util.*;

public class Mic1Machine extends Mic1Kernel implements Mic1Interface,PulseConsumer
{
	private IJVMLoader m_Program = null;
	private Mic1OutputInterface m_Output = null;
	private PulseGenerator m_Clock = null;
	private boolean m_RunToBreakpoint = false;
	private BreakPoints m_BreakPoints = new BreakPoints();

	public Mic1Machine()
	{
		super();
		try
		{
			String sPath = new File(".").getAbsolutePath();
			sPath = sPath.substring(0,sPath.length()-1);
			sPath += "mic1ijvm.mic1";
			loadMicroprogram(sPath);
		}
		catch(Exception exc)
		{
			System.out.println(exc);
		}
	}

	/////////////////////////////////////////////////////////////////////
	// Clock
	/////////////////////////////////////////////////////////////////////
	public boolean OnClock() throws Exception
	{
		assertValidMic();
		boolean bReturn = cycle();
		refreshOutput();

		if(m_RunToBreakpoint && getBreakpointHit()>=0)
		{
			bReturn = false;
		}
		return bReturn;
	}

	public void OnStart(){}

	public void OnStop()
	{
		if(m_Output != null)
		{
			try
			{
				int nLine = getCurrentProgramLineNr();
				m_Output.OnStop(nLine, getCycleCount());
				nLine = getBreakpointHit();
				if(m_RunToBreakpoint && nLine>=0)
					m_Output.OnReachedBreakPoint(nLine);
			}
			catch(Exception e){}
		}
	}

	/**
	 * Loads a new program in memory.
	 *
	 * @param : filepath of the program
	 * @exception : MicException if there is still an active program.
	 * @exception : IOException for file access;
	 */
	public void loadProgram (String sFileName) throws Exception
	{
		assertChangeMic();

		try
		{
			m_Program = new IJVMLoader(sFileName);
		}
		catch(Exception e)
		{
			m_Output.OnChangedStandardOut("No IJVM file Found");
		}
		if(m_Program.isValid() == null)
		{
			setMemory(m_Program.getProgram());
			refreshOutput();
		}
		else if(m_Output != null)
		{
			m_Output.OnChangedStandardOut("Error loading program : "+m_Program.isValid()+"\n");
			setMemory(null);
			m_Program = null;
		}
		reset();
	}


	/**
	 * Returns true when there is valid program running/loaded.
	 *
	 * @return : boolean!
	 */
	public boolean validProgram ()
	{
		return validMemory() && m_Program != null;
	}

	/**
	 * Returns the filepath of the active program file
	 *
	 * @return : Empty String when there is no program active.
	 */
	public String getProgramFile ()
	{
		return m_Program != null ? m_Program.getFileName():"";
	}


	/**
	 * Start or Resume the current program. Breakpoints are skipped
	 *
	 * @exception : MicException if there is no program loaded.
	 * @exception : MicException if there is a running program.
	 */
	public void run () throws Exception
	{
		m_RunToBreakpoint = false;
		startMachine();
	}


	/**
	 * Start or Resume the current program until the next breakpoint.
	 *
	 * @exception : MicException if there is no program loaded.
	 * @exception : MicException if there is a running program.
	 */
	public void runToBreakPoint () throws Exception
	{
		m_RunToBreakpoint = true;
		startMachine();
	}

	/**
	 * Start or Resumes the current program
	 *
	 * @exception : MicException if there is no program loaded.
	 * @exception : MicException if there is a running program.
	 */
	private void startMachine() throws Exception
	{
		assertValidMic();
		assertChangeMic();
		if(m_Clock == null)
			m_Clock = new PulseGenerator(this, 1);
		if(m_Output != null)
			m_Output.OnRun();
		m_Clock.startThread();
	}

	/**
	 * Executes the current instruction and increments the PC.
	 *
	 * @return : Returns false when end of program is reached.
	 * @exception : MicException if there is no program loaded.
	 * @exception : MicException if there is a running program.
	 */
	public boolean step () throws Exception
	{
		assertValidMic();
		assertChangeMic();
		boolean bReturn = cycle();
		refreshOutput();
		return bReturn;
	}


	/**
	 * Resets the machine. The program loaded persists in memory
	 *
	 * @exception : MicException if there is a running program.
	 */
	public void reset () throws Exception
	{
		assertChangeMic();
		super.reset();
		refreshOutput();
	}

	/**
	 * Forces a Kernal Resets of the machine. The program loaded is cleared from memory
	 *
	 * @exception : MicException if there is a running program.
	 */
	public void unloadProgram () throws Exception
	{
		m_Program = null;
		assertChangeMic();
		super.forceReset();
		refreshOutput();
	}

	/**
	 * Stops the current program if running.
	 *
	 */
	public void stop ()
	{
		if(m_Clock != null)
		{
	    	m_Clock.stopThread();
	    	m_Clock = null;
	    }
	}

	/**
	 * Determines if the program is running.
	 *
	 * @return : true when running!
	 */
	public boolean isRunning ()
	{
		if(m_Clock != null)
	    	return m_Clock.isRunning();
	    return false;
	}

	/**
	 * Adds a new linenumber to list overlapping the old one.
	 *
	 * @param : the linenumber of the breakpoint.
	 * @exception : MicException if there is a running program.
	 */
	public void setBreakPoint (int nLineNumber) throws Exception
	{
		assertChangeMic();
		m_BreakPoints.setBreakPoint(nLineNumber);
	}

	/**
	 * Removes a linenumber from the list if found.
	 *
	 * @param : the linenumber of the breakpoint.
	 * @exception : MicException if there is a running program.
	 * @return : false if the breakpoint doesn't exit
	 */
	public boolean removeBreakPoint (int nLineNumber) throws Exception
	{
		assertChangeMic();
		return m_BreakPoints.removeBreakPoint(nLineNumber);
	}

	/**
	 * Validates the linenumbers of the breakpoint list.
	 * Removing and returning the invalid ones.
	 *
	 * @exception : MicException if there is no program loaded.
	 * @exception : MicException if there is a running program.
	 */
	public int[] validateBreakPoints () throws Exception
	{
		assertValidMic();
		assertChangeMic();
		int[] nArray = getBreakPoints();
		clearBreakPoints();
		Hashtable lineTable = m_Program.getLineNumberTable();
		if(lineTable != null)
		{
			Enumeration enum = lineTable.elements();
			Integer integer = (Integer)enum.nextElement();
			while(integer != null)
			{
				for(int i=0; i<nArray.length; i++)
				{
					if(integer.intValue() == nArray[i])
					{
						setBreakPoint(nArray[i]);
						break;
					}
				}
				integer = (Integer)enum.nextElement();
			}
		}
		return getBreakPoints();
	}

	/**
	 * Removes all breakpoints.
	 *
	 * @exception : MicException if there is a running program.
	 */
	public void clearBreakPoints () throws Exception
	{
		assertChangeMic();
		m_BreakPoints.clear();
	}

	/**
	 * Fills and returns an array containing the breakpoint.
	 * It's not a reference to the internal list.
	 *
	 * @return : int array containing the linenumbers.
	 */
	public int[] getBreakPoints ()
	{
		return m_BreakPoints.toIntArray();
	}

	/**
	 * The interface which is used by the machine for output messages
	 *
	 */
	public void setOutputInterface (Mic1OutputInterface output)
	{
		m_Output = output;
	}


	/////////////////////////////////////////////////////////////////////
	// Setting register values
	/////////////////////////////////////////////////////////////////////

	/**
	 * Memory address register.
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterMAR (int nValue) throws Exception
	{
		assertChangeMic();
		mar.forceValue(nValue);
		refreshOutput();
	}

	/**
	 * Register that contains word values for writing or reading
	 * from main memory
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterMDR (int nValue) throws Exception
	{
		assertChangeMic();
		mdr.forceValue(nValue);
		refreshOutput();
	}

	/**
	 * Program Counter
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterPC (int nValue) throws Exception
	{
		assertChangeMic();
		pc.forceValue(nValue);
		refreshOutput();
	}

	/**
	 * Register that stores bytes fetched from memory.
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterMBR (int nValue) throws Exception
	{
		assertChangeMic();
		mbr.forceValue(nValue);
		refreshOutput();
	}

	/**
	 * Top point of the current stackframe
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterSP (int nValue) throws Exception
	{
		assertChangeMic();
		sp.forceValue(nValue / 4);
		refreshOutput();
	}

	/**
	 * Bottom point of the current stackframe
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterLV (int nValue) throws Exception
	{
		assertChangeMic();
		lv.forceValue(nValue / 4);
		refreshOutput();
	}

	/**
	 * Points to the const pool
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterCPP (int nValue) throws Exception
	{
		assertChangeMic();
		cpp.forceValue(nValue / 4);
		refreshOutput();
	}

	/**
	 * ??
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterTOS (int nValue) throws Exception
	{
		assertChangeMic();
		tos.forceValue(nValue);
		refreshOutput();
	}

	/**
	 * scratch register
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterOPC (int nValue) throws Exception
	{
		assertChangeMic();
		opc.forceValue(nValue);
		refreshOutput();
	}

	/**
	 * ??
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterH (int nValue) throws Exception
	{
		assertChangeMic();
		h.forceValue(nValue);
		refreshOutput();
	}

	/**
	 * Contains the address of the next microinstruction.
	 *
	 * @param : Value.
	 * @exception : MicException if there is a running program.
	 */
	public void setRegisterMPC (int nValue) throws Exception
	{
		assertChangeMic();
		mpc.forceValue(nValue);
		refreshOutput();
	}

	/**
	 * Sends a character to the machines keyboard buffer.
	 *
	 * @param : a character.
	 */
	public void sendChar(char c)
	{
		getKeyBuffer().addChar(c);
	}



	public void writeCharToConsole(char c)
	{
		m_Output.OnCharOut(c);
	}

	/**
	 * Returns the byte at the specified memory address.
	 * If the address is out of range, this method
	 * returns 255 (0xFF), like in most real computer
	 * systems.
	 *
	 * @param : memory address.
	 */
	public int getMemoryContent(int address)
	{
		return main_memory.getByteAt(address);
	}


	/////////////////////////////////////////////////////////////////////
	// Output
	/////////////////////////////////////////////////////////////////////
	public void refreshOutput()
	{
		if(m_Output != null)
		{
			int nLine = 0;
			try{nLine = getCurrentProgramLineNr();}catch(Exception e){}

			m_Output.OnChangedDebugOut("");
			m_Output.OnChangedMAR(mar.getValue() * 4);
			m_Output.OnChangedMDR(mdr.getValue());
			m_Output.OnChangedPC(pc.getValue());
			m_Output.OnChangedMBR(mbr.getValue());
			m_Output.OnChangedLV(lv.getValue());
			m_Output.OnChangedCPP(cpp.getValue());
			m_Output.OnChangedTOS(tos.getValue());
			m_Output.OnChangedOPC(opc.getValue());
			m_Output.OnChangedH(h.getValue());
			m_Output.OnChangedMIR(mir.toString(), mpc.getNextInstruction());
			m_Output.OnChangedMPC(mpc.getValue());
			m_Output.OnChangedALU(alu.getN(), alu.getZ(), alu.getOut());
			m_Output.OnChangedSP(sp.getValue());
			m_Output.OnStep(nLine, getCycleCount());
			//m_Output.OnChangedStandardOut(String sText);
			//m_Output.OnChangedDebugOut(String sText);
		}
	}
	/////////////////////////////////////////////////////////////////////
	// Helpers
	/////////////////////////////////////////////////////////////////////
	protected int getCurrentProgramLineNr() throws Exception
	{
		assertValidMic();
		Hashtable lineTable = m_Program.getLineNumberTable();
		if(lineTable != null)
		{
			int instrValue = getProgramCounterValue();
			if(instrValue != -1)
			{
				Integer line = (Integer)lineTable.get(new Integer(instrValue));
				if(line != null)
					return line.intValue();
			}
		}
		return -1;
	}

	protected int getBreakpointHit() throws Exception
	{
		int[] nArray = getBreakPoints();
		int nLine = getCurrentProgramLineNr();
		for(int i=0; i<nArray.length; i++)
		{
			if(nArray[i] == nLine)
			{
				return nLine;
			}
		}
		return -1;
	}

	protected void assertChangeMic() throws Exception
	{
		if(isRunning())
			throw new MicException("Error: ASSERT CHANGING MIC VALUE\n");
	}

	protected void assertValidMic() throws Exception
	{
		if(!validProgram())
			throw new MicException("Error: NO VALID MEMORY\n");
	}
}
